/*
 * Rdriver.c
 * 
 * author: Hector Corrada Bravo (hcorrada@gmail.com)
 * created: Aug 11, 2009
 *
 * R interface to spch reader
 */

#include <R.h>
#include <Rinternals.h>

#include "spchreader.h"
#define NMSIZE 255

/*
 * The spch reader does things in two cycles, it first gets information
 * about the file, e.g. number of panels, number of primer cycles,
 * number of ligation cycles, number of beads, etc., usign function
 * getspchinfo.
 *
 * This information is used to populate a struct of type fileInfo.
 * Once this is done, that information is used to create a container for the
 * file data, the function init_fileData does exactly that.
 * 
 * Once the container of type fileData is created it is passed to the readspch function
 * to be populated with file data.
 *
 * The function pack_result makes the fileData container into something that can
 * be passed back to R
 *
 * The first slot in the result is a status flag
 * 0 indicates success, < 0 indicates failure
 */
static void init_fileData (const fileInfo *file_info, fileData *file_data);
static SEXP pack_result (fileData *file_data);
SEXP rdriver (SEXP pfilename, SEXP Rverbose);

/*
 * void init_fileData (const fileInfo *file_info, fileData *file_data)
 * 
 * Instantiates a container of type fileData to be populated according to the
 * file information stored in file_info
 *
 * It does not flag errors, but R_alloc could make it bonk out
 * 
 * args:
 *   file_info: structure with information about spch file
 *   file_data: container to be instantiated
 *
 * returns: void
 */
static void init_fileData (const fileInfo *file_info, fileData *file_data)
{
  int npanels;
  int nprimer_cycles;
  int nligation_cycles;
  int nbeads;
  int i,j;
  panelData *panel_data;
  panelInfo *panel_info;

  npanels = file_data->npanels = file_info->npanels;
  file_data->filename = file_info->filename;

  /* allocate space for panel_data objects */
  file_data->panel_data = (panelData *) R_alloc(npanels, sizeof(panelData));
  
  /* allocate space in panel_data objects */
  for (i=0; i<npanels; i++) {
    panel_data = file_data->panel_data+i;
    panel_info = file_info->panel_info+i;
    
    /* copy panel information */
    memcpy(panel_data->name, panel_info->name, strlen(panel_info->name));
    nprimer_cycles = panel_data->nprimer_cycles = panel_info->nprimer_cycles;
    panel_data->nligation_cycles = (int *) R_alloc(nprimer_cycles, sizeof(int));

    for (j=0; j<nprimer_cycles; j++) {
      panel_data->nligation_cycles[j] = panel_info->nligation_cycles[j];
    }
    nligation_cycles = panel_data->total_nligation_cycles = panel_info->total_nligation_cycles;
    nbeads = panel_data->nbeads = panel_info->nbeads;

    /* allocate space for  primer cycle names */
    panel_data->primer_names = (char **) R_alloc(nprimer_cycles,  sizeof(char *));
    for (j=0; j<nprimer_cycles; j++) {
      panel_data->primer_names[j] = (char *) R_alloc(NMSIZE, sizeof(char));
    }

    /* allocate space for ligation cycle names */
    panel_data->ligation_names = (char **) R_alloc(nligation_cycles, sizeof(char *));
    for (j=0; j<nligation_cycles; j++) {
      panel_data->ligation_names[j] = (char *) R_alloc(NMSIZE, sizeof(char));
    }

    #if DEBUG
    printf("Ready to allocate panel\n");        
    #endif

    /* allocate space for intensities */
    panel_data->intensities = (float **) R_alloc(nligation_cycles, sizeof(float *));
    for (j=0; j<nligation_cycles; j++) {
      panel_data->intensities[j] = (float *) R_alloc(nbeads * 4, sizeof(float));
    }

    /* allocate space for color calls */
    panel_data->color_calls = (char **) R_alloc(nligation_cycles, sizeof(char *));
    for (j=0; j<nligation_cycles; j++) {
      panel_data->color_calls[j] = (char *) R_alloc(nbeads, sizeof(char));
    }

    #if DEBUG
    printf("Ready to allocate panel 2\n");        
    #endif

    /* allocate space for yxloc */
    panel_data->yxloc = (int *) R_alloc(nbeads * 2, sizeof(int));
  }
}

/*
 * SEXP pack_result (int **yxloc, fileData *file_data)
 *
 * Packs the file data (stored in file_data) along with the
 * yxlocation of each panel into a list that can be passed to R
 *
 * args: 
 *    yxloc: array of int[], of length file_data->npanels
 *           each int[] is of size file_data->panel_data[i].nbeads by 2
 *    file_data: data read from spch file
 *
 * returns list SEXP of length file_data->npanels + 1with the following structure
 *     - vector of panel names
 *     - info list for panel 1
 *         -- character vector with primer cycle names
 *         -- character vector with ligation cycle names
 *         -- int vector with yx locations
 *         -- list with intensity values (each component is one cycle of intensities)
 *         -- list with color calls (each component is one cycle of color calls)
 *    - info list for panel 2
 *    - etc.
 */
static SEXP pack_result (fileData *file_data)
{
  int i, j, k;

  SEXP result;
  SEXP panel_names;
  SEXP p_panel_data;
  SEXP yxLocation;
  SEXP primer_names;
  SEXP ligation_names;
  SEXP nligation_cycles;
  SEXP intensities, cycle_ints;
  SEXP color_calls, cycle_calls;

  double *re_intensities;
  int *int_calls;

  panelData panel_data;

  /* allocate result list, one slot per panel and one slot for panel names vector */
  PROTECT(result = allocVector(VECSXP, file_data->npanels+1));

  /* allocate panel names vector */
  PROTECT(panel_names = allocVector(STRSXP, file_data->npanels));

  /* pack each panel's result */
  for (i=0; i<file_data->npanels; i++) {

    /* 
     * allocate panel data list of size 6:
     *   1) primer cycle names character vector
     *   2) ligation cycle names character vector
     *   3) number of ligation cycles in each primer cycle
     *   4) yx locations int vector
     *   5) intensities double vector
     *   6) color calls int vector
     */
    PROTECT(p_panel_data = allocVector(VECSXP, 6));
    panel_data = file_data->panel_data[i];

    /* get this panel's name */
    SET_STRING_ELT(panel_names, i, mkChar(panel_data.name));

    /* grab the primer cycle names */
    PROTECT(primer_names = allocVector(STRSXP, panel_data.nprimer_cycles));
    for (j=0; j<panel_data.nprimer_cycles; j++) {
      SET_STRING_ELT(primer_names, j, mkChar(panel_data.primer_names[j]));
    }
    SET_VECTOR_ELT(p_panel_data, 0, primer_names);

    /* grab the ligation cycle names */
    PROTECT(ligation_names = allocVector(STRSXP, panel_data.total_nligation_cycles));
    for (j=0; j<panel_data.total_nligation_cycles; j++) {
      SET_STRING_ELT(ligation_names, j, mkChar(panel_data.ligation_names[j]));
    }
    SET_VECTOR_ELT(p_panel_data, 1, ligation_names);

    /* grab the number of ligation cycles in each primer cycle vector */
    PROTECT(nligation_cycles = allocVector(INTSXP, panel_data.nprimer_cycles));
    memcpy(INTEGER(nligation_cycles), panel_data.nligation_cycles, panel_data.nprimer_cycles * sizeof(int));
    SET_VECTOR_ELT(p_panel_data, 2, nligation_cycles);

    /* grab the yx location vector */
    PROTECT(yxLocation = allocMatrix(INTSXP, panel_data.nbeads, 2));
    memcpy(INTEGER(yxLocation), panel_data.yxloc, panel_data.nbeads * 2 * sizeof(int));
    SET_VECTOR_ELT(p_panel_data, 3, yxLocation);

    /* grab panel intensities */
    PROTECT(intensities = allocVector(VECSXP, panel_data.total_nligation_cycles));
    for (j=0; j<panel_data.total_nligation_cycles; j++) {
      PROTECT(cycle_ints = allocVector(REALSXP, panel_data.nbeads * 4));
      re_intensities = REAL(cycle_ints);
      for (k=0; k<(panel_data.nbeads * 4); k++) {
	re_intensities[k] = (double) panel_data.intensities[j][k];
      }
      SET_VECTOR_ELT(intensities, j, cycle_ints);
      UNPROTECT(1);
    }
    SET_VECTOR_ELT(p_panel_data, 4, intensities);

    /* grab panel color calls */
    PROTECT(color_calls = allocVector(VECSXP, panel_data.total_nligation_cycles));
    for (j=0; j<panel_data.total_nligation_cycles; j++) {
      PROTECT(cycle_calls = allocVector(INTSXP, panel_data.nbeads));
      int_calls = INTEGER(cycle_calls);
      for (k=0; k<(panel_data.nbeads); k++) {
	int_calls[k] = (int) panel_data.color_calls[j][k];
      }
      SET_VECTOR_ELT(color_calls, j, cycle_calls);
      UNPROTECT(1);
    }
    SET_VECTOR_ELT(p_panel_data, 5, color_calls);

    /* store this panel's data in result list */
    SET_VECTOR_ELT(result, i+1, p_panel_data);
    UNPROTECT(7);
  }
  SET_VECTOR_ELT(result, 0, panel_names);
  UNPROTECT(2);
  return result;
 }

/*
 * SEXP rdriver (SEXP pfilename)
 *
 * Entry point from R
 *
 * args:
 *    pfilename: spch file name
 *
 * returns SEXP list of length 2 with structure
 *  - status: status flag returned by readspch function
 *  - resData: SEXP returned by function pack_result
 *
 */  
SEXP rdriver (SEXP pfilename, SEXP Rverbose)
{
  const char *filename;
  fileInfo file_info;
  fileData file_data;
  panelInfo panel_info;

  int status, i;
  SEXP p_status;
  SEXP primer_names;
  SEXP ligation_names;
  SEXP p_intensities;
  SEXP p_yxloc;
  SEXP resData;
  SEXP result;

  int verbose = INTEGER(Rverbose)[0];

  PROTECT(p_status = allocVector(INTSXP, 1));

  filename = CHAR(STRING_ELT(pfilename, 0));
  status = (int) getspchinfo(filename, &file_info, verbose);

  if (status >= 0) {
    if (verbose)
      print_fileInfo(&file_info);

    init_fileData(&file_info, &file_data);
    free_fileInfo(&file_info);

    status = (int) readspch(filename, &file_data, verbose);
  }

  if (verbose)
    Rprintf("Packing result for R\n");

  if (status >= 0) {
    if (verbose)
      print_fileData(&file_data);

    PROTECT(resData = pack_result(&file_data));
  } else {
    PROTECT(resData = allocVector(NILSXP, 1));
  }

  INTEGER(p_status)[0] = status;
  PROTECT(result = allocVector(VECSXP, 2));
  SET_VECTOR_ELT(result, 0, p_status);
  SET_VECTOR_ELT(result, 1, resData);

  UNPROTECT(3);
  return (result);
}
  
