/*
 * reader.c
 * author: Hector Corrada Bravo (hcorrada@gmail.com)
 * created: Aug 11, 2009
 *
 * Functions to read data from an spch file
 *
 */
#include "spchreader.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>

#define BUFFERSIZE 1024

/*
 * call-back function used at each panel data node
 * gets information about the panel node being visited
 */
static herr_t panelIter_info (hid_t g_id, const char *name, const H5L_info_t *info, void *op_data)
{
  hid_t me;
  hid_t f3;
  hid_t ref_data;
  hid_t dataset;
  hid_t filespace;
  //  hid_t memspace;

  int i;

  hsize_t dims[2];
  
  H5G_info_t group_info;
  herr_t status;

  fileInfo *file_info; 
  panelInfo *panel_info; 

  file_info = (fileInfo *) op_data;
  panel_info = file_info->panel_info+file_info->next++;

  #ifdef DEBUG
  printf("Visiting panel %s\n", name);
  #endif

  /* storepanel name */
  memcpy(&(panel_info->name), name, strlen(name));
  panel_info->name[strlen(name)] = '\0';

  me = H5Gopen(g_id, name, H5P_DEFAULT);
  if (me < 0) {
    printf("Error in panel info iterator %s: couldn't open group\n", name);
    H5Eprint2(H5E_DEFAULT, NULL);
    return -1;
  }

  f3 = H5Gopen(me, "F3", H5P_DEFAULT);
  if (f3 < 0) {
    printf("Error in panel info iterator %s: couldn't open F3 group\n", name);
    H5Eprint2(H5E_DEFAULT, NULL);
    H5Gclose(me);
    return -1;
  }

  status = H5Gget_info(f3, &group_info);
  if (status < 0) {
    printf("Error in panel info iterator %s: couldn't get F3 group info\n", name);
    H5Eprint2(H5E_DEFAULT, NULL);
    H5Gclose(f3);
    H5Gclose(me);
    return status;

  }

  /* store number of primer cycles */
  panel_info->nprimer_cycles = group_info.nlinks;
  if (panel_info->nprimer_cycles != NPRIMER_CYCLES) {
    printf("Number of primer cycles in panel %s: expected %d but there are %d\n", name, NPRIMER_CYCLES, panel_info->nprimer_cycles);
  }
  panel_info->nligation_cycles = malloc(panel_info->nprimer_cycles * sizeof(int));

  panel_info->total_nligation_cycles = 0;
  for (i=0; i < panel_info->nprimer_cycles; i++) {
    status = H5Gget_info_by_idx(f3, ".", H5_INDEX_NAME, H5_ITER_INC, i, &group_info, H5P_DEFAULT);
    if (status < 0) {
      printf("Error getting number of ligation cycles in primer cycle %d, panel %s\n", i, name);
      H5Eprint2(H5E_DEFAULT, NULL);
      H5Gclose(f3);
      H5Gclose(me);
      return status;
      
    }
    /* store number of ligation cycles */
    panel_info->nligation_cycles[i] = group_info.nlinks;
    panel_info->total_nligation_cycles += group_info.nlinks;
  }
  H5Gclose(f3);

  /* grab the number of beads */
  ref_data = H5Gopen(me, "RefData/BeadsP2", H5P_DEFAULT);
  if (ref_data < 0) {
    printf("Error in panel info iterator %s: couldn't open beads group\n", name);
    H5Eprint2(H5E_DEFAULT, NULL);
    H5Gclose(me);
    return -1;
  }
  dataset = H5Dopen(ref_data, "yxLocation", H5P_DEFAULT);
  if (dataset < 0) {
    printf("Error in panel info iterator %s: couldn't open location group\n", name);
    H5Eprint2(H5E_DEFAULT, NULL);
    H5Gclose(ref_data);
    H5Gclose(me);
    return -1;
  }
  
  filespace = H5Dget_space(dataset);
  if (filespace < 0) {
    printf("Error in panel info iterator %s: couldn't get location filespace\n", name);
    H5Eprint2(H5E_DEFAULT, NULL);
    H5Dclose(dataset);
    H5Gclose(ref_data);
    H5Gclose(me);
    return -1;
  }

  status = H5Sget_simple_extent_dims(filespace, dims, NULL);
  if (status < 0) {
    printf("Error in panel info iterator %s: couldn't get location array dimensions\n", name);
    H5Eprint2(H5E_DEFAULT, NULL);
    H5Sclose(filespace);
    H5Dclose(dataset);
    H5Gclose(ref_data);
    H5Gclose(me);
    return status;
  }

  panel_info->nbeads = dims[1];
  H5Sclose(filespace);
  H5Dclose(dataset);
  H5Gclose(ref_data);
  H5Gclose(me);
  return 0;
}

/*
 * get information about an spch file, stored in given fileInfo structure
 */
herr_t getspchinfo(const char *filename, fileInfo *file_info, int verbose)
{
  hid_t file;
  hid_t panels;
  H5G_info_t group_info;
  herr_t status;

  status = H5Eset_auto2(H5E_DEFAULT, NULL, NULL);
  if (status < 0) {
    printf("Error in getspchinfo: couldn't setup error reproter\n");
    H5Eprint2(H5E_DEFAULT, NULL);
    return status;
  }

  if (verbose)
    printf("Reading spch file %s\n", filename);

  memcpy(file_info->filename, filename, strlen(filename));

  file = H5Fopen(filename, H5F_ACC_RDONLY, H5P_DEFAULT);
  if (file < 0) {
    printf("Error in getspchinfo: couldn't open file %s\n", filename);
    H5Eprint2(H5E_DEFAULT, NULL);
    return -1;
  }
    
  panels = H5Gopen(file,"Panels",H5P_DEFAULT);
  if (panels < 0) {
    printf("Error in getspchinfo: couldn't get panels group\n");
    H5Eprint2(H5E_DEFAULT, NULL);
    H5Fclose(file);
    return -1;
  }

  status = H5Gget_info(panels, &group_info);
  if (status < 0) {
    printf("Error in getspchinfo: couldn't get Panels group info\n");
    H5Eprint2(H5E_DEFAULT, NULL);
    H5Gclose(panels);
    H5Fclose(file);
    return status;
  }

  if (verbose)
    printf("Number of panels %d\n", (int) group_info.nlinks);

  /* store the number of panels in file */
  file_info->npanels = (int) group_info.nlinks;
  
  file_info->panel_info = malloc(file_info->npanels * sizeof(panelInfo));
  file_info->next = 0;

  /* iterate over panel data nodes */
  status = H5Literate (panels, H5_INDEX_NAME, H5_ITER_INC, NULL, &panelIter_info, file_info);
  H5Gclose(panels);
  H5Fclose(file);
  return status;
}

/*
 * helper function reads dataset in group gid with name name and into buf 
 */
static herr_t my_read_dataset (hid_t gid, const char *name, const hid_t output_type, void *buf)
{
  hid_t dataset, filespace, memspace, datatype;
  hsize_t *dims;
  int nelmts;
  int rank;

  int i;
  herr_t status;
  
  dataset = H5Dopen(gid, name, H5P_DEFAULT);
  if (dataset < 0) {
    printf("Error in my_read_dataset: couldn't open dataset\n");
    H5Eprint2(H5E_DEFAULT, NULL);
    return -1;
  }

  datatype = H5Dget_type(dataset);
  if (datatype < 0) {
    printf("Error in my_read_dataset: couldn't get datatype\n");
    H5Eprint2(H5E_DEFAULT, NULL);
    H5Dclose(dataset);
    return -1;
  }

  filespace = H5Dget_space(dataset);
  if (filespace < 0) {
    printf("Error in my_read_dataset: couldn't get filespace\n");
    H5Eprint2(H5E_DEFAULT, NULL);
    H5Tclose(datatype);
    H5Dclose(dataset);
    return -1;
  }

  rank = H5Sget_simple_extent_ndims(filespace);
  dims = malloc(rank * sizeof(hsize_t));

  status = H5Sget_simple_extent_dims(filespace, dims, NULL);
  if (status < 0) {
    printf("Error in my_read_dataset: couldn't get dimensions\n");
    H5Eprint2(H5E_DEFAULT, NULL);
    free(dims);
    H5Sclose(filespace);
    H5Tclose(datatype);
    H5Dclose(dataset);
    return status;
  }

  #ifdef DEBUG
  for (i=0; i<rank; i++) 
    printf("%d ", (int) dims[i]);
  printf("\n");
  #endif

  memspace = H5Screate_simple(rank, dims, NULL);
  if (memspace < 0) {
    printf("Error in my_read_dataset: couldn't create memspace\n");
    H5Eprint2(H5E_DEFAULT, NULL);
    free(dims);
    H5Sclose(filespace);
    H5Tclose(datatype);
    H5Dclose(dataset);
    return -1;
  }

 nelmts = 1;
 for (i=0; i<rank; i++)
   nelmts *= dims[i];
 free(dims);

 status = H5Dread(dataset, datatype, memspace, filespace, H5P_DEFAULT, buf);
 if (status < 0) {
   printf("Error in my_read_dataset: couldn't read data\n");
   H5Eprint2(H5E_DEFAULT, NULL);
   H5Sclose(memspace);
   H5Sclose(filespace);
   H5Tclose(datatype);
   H5Dclose(dataset);
   return status;
 }
 H5Sclose(memspace);
 H5Sclose(filespace);

 status = H5Tconvert(datatype, output_type, nelmts, buf, NULL, H5P_DEFAULT);
 if (status < 0) {
   printf("Error in my_read_dataset: couldn't convert datatypes\n");
 }

 H5Tclose(datatype);
 H5Dclose(dataset);

 return status;
} 

/*
 * call-back function used in each ligation cycle data node
 * intensities are read here
 */
static herr_t ligationIter_data (hid_t g_id, const char *name, const H5L_info_t *info, void *op_data)
{
  hid_t me;
  herr_t status;

  int offset;
  panelData *panel_data;
  int ligation_pos;

  panel_data = (panelData *) op_data;

  #ifdef DEBUG
  printf("Getting data from ligation %s\n", name);
  #endif

/*   primer_pos = panel_data->next_primer; */
/*   offset = panel_data->next_primer * panel_data->nligation_cycles[0] + panel_data->next_ligation; */
/*   assert(offset <= (panel_data->nligation_cycles[0] * panel_data->nprimer_cycles)); */
  
  ligation_pos = panel_data->next_ligation++;

  /* store the name of this ligation cycle */
  memcpy(panel_data->ligation_names[ligation_pos], name, strlen(name));
  panel_data->ligation_names[ligation_pos][strlen(name)] = '\0';

  me = H5Gopen(g_id, name, H5P_DEFAULT);
  if (me < 0) {
    printf("Error in ligation cycle iterator: couldn't open group\n");
    H5Eprint2(H5E_DEFAULT, NULL);
    return -1;
  }

  /* read intensities into intensity array in panel_data */
  status = my_read_dataset(me, "scaledBeads", H5T_NATIVE_FLOAT, (void *) panel_data->intensities[ligation_pos]);
  if (status < 0) {
    printf("Error in ligation cycle iterator: couldn't read intensity data\n");
    H5Eprint2(H5E_DEFAULT, NULL);
    H5Gclose(me);
    return status;
  }

  /*read color calls into color calls array in panel_data */
  status = my_read_dataset(me, "colorCall", H5T_NATIVE_CHAR, (void *) panel_data->color_calls[ligation_pos]);
  if (status < 0) {
    printf("Error in ligation cycle iterator: couldn't read color calls\n");
    H5Eprint2(H5E_DEFAULT, NULL);
    H5Gclose(me);
    return status;
  }
  return 0;
}

/*
 * call-back function used in each primer cycle data node
 *  iterates over ligation cycles in children nodes
 */
static herr_t primerIter_data (hid_t g_id, const char *name, const H5L_info_t *info, void *op_data)
{
  hid_t me;
  herr_t status;
  panelData *panel_data;
  int primer_pos;
  
  H5G_info_t group_info;

  panel_data = (panelData *) op_data;

  #ifdef DEBUG
  printf("Getting data from primer %s\n", name);
  #endif

  /* get primer cycle name if this is the first time we see it */
  primer_pos = panel_data->next_primer;
  memcpy(panel_data->primer_names[primer_pos],name,strlen(name));
  panel_data->primer_names[primer_pos][strlen(name)] = '\0';

  me = H5Gopen(g_id, name, H5P_DEFAULT);
  if (me < 0) {
    printf("Error in primer cycle iterator: couldn't open group\n");
    H5Eprint2(H5E_DEFAULT, NULL);
    return -1;
  }

  status = H5Gget_info(me, &group_info);
  if (status < 0) {
    printf("Error in primer cycle iterator: couldn't get group info\n");
    H5Eprint2(H5E_DEFAULT, NULL);
    H5Gclose(me);
    return status;
  }
  if (group_info.nlinks != panel_data->nligation_cycles[primer_pos]) {
    printf("Unexpected number of ligation cycles: expected %d, saw %d\n", panel_data->nligation_cycles[primer_pos], (int) group_info.nlinks);
    H5Gclose(me);
    return -1;
  }

  #ifdef DEBUG
  printf("Ready to iterate over ligations\n");
  #endif

  /* iterate over ligation cycles */
  status = H5Literate (me, H5_INDEX_NAME, H5_ITER_INC, NULL, &ligationIter_data, panel_data);
  H5Gclose(me);
  panel_data->next_primer++;
  return status;
}

/*
 * call-back function used on each panel data node
 * iterates over primer cycles in children nodes 
 */
static herr_t panelIter_data (hid_t g_id, const char *name, const H5L_info_t *info, void *op_data)
{
  hid_t me, f3, ref_data;
  herr_t status;
  
  fileData *file_data;
  panelData *panel_data;

  file_data = (fileData *) op_data;
  panel_data = file_data->panel_data+file_data->next++;

  #ifdef DEBUG
  printf("Getting data from panel %s\n", name);
  #endif

  me = H5Gopen(g_id, name, H5P_DEFAULT);
  if (me < 0) {
    printf("Error in panel iterator: couldn't open group\n");
    H5Eprint2(H5E_DEFAULT, NULL);
    return -1;
  }

  /* grab the yx locations */
  ref_data = H5Gopen(me, "RefData/BeadsP2", H5P_DEFAULT);
  if (ref_data < 0) {
    printf("Error in panel iterator: couldn't get bead locations group\n");
    H5Eprint2(H5E_DEFAULT, NULL);
    H5Gclose(me);
    return -1;
  }
  status = my_read_dataset(ref_data, "yxLocation", H5T_NATIVE_INT, (void *) panel_data->yxloc);
  if (status < 0) {
    printf("Error in panel iterator: couldn't get bead location data\n");
    H5Eprint2(H5E_DEFAULT, NULL);
    H5Gclose(ref_data);
    H5Gclose(me);
    return status;
  }

  
  f3 = H5Gopen(me, "F3", H5P_DEFAULT);
  if (f3 < 0) {
    printf("Error in panel iterator: couldn't open F3 group\n");
    H5Eprint2(H5E_DEFAULT, NULL);
    H5Gclose(me);
    return -1;
  }

  
  /* iterate over primer cycles */
  panel_data->next_primer = 0;
  panel_data->next_ligation = 0;
  status = H5Literate (f3, H5_INDEX_NAME, H5_ITER_INC, NULL, &primerIter_data, panel_data);
  return status;
}

/*
 * populate fileData structure with data from spch file
 * iterates over panel data nodes
 */
herr_t readspch (const char *filename, fileData *file_data, int verbose)
{
  hid_t file;
  hid_t panels;

  herr_t status;

  status = H5Eset_auto2(H5E_DEFAULT, NULL, NULL);
  if (status < 0) {
    printf("Error in readspch: couldn't set error printer\n");
    H5Eprint2(H5E_DEFAULT, NULL);
    return status;
  }
  if (verbose)
    printf("Reading data from spch file: %s\n", filename);

  file = H5Fopen(filename, H5F_ACC_RDONLY, H5P_DEFAULT);
  if (file < 0) {
    printf("Error in readspch: couldn't open file\n");
    H5Eprint2(H5E_DEFAULT, NULL);
    return -1;
  }
    
  panels = H5Gopen(file, "Panels", H5P_DEFAULT);
  if (panels < 0) {
    printf("Error in readspch: couldn't open panels group\n");
    H5Eprint2(H5E_DEFAULT, NULL);
    H5Fclose(file);
    return -1;
  }
  
  /* iterate over panels */
  file_data->next=0;
  status = H5Literate (panels, H5_INDEX_NAME, H5_ITER_INC, NULL, &panelIter_data, file_data);
  return status;
}
